from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

if TYPE_CHECKING:
    import sys
    from re import Pattern
    from types import TracebackType

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack

    from .httpserver import HTTPServer
    from .httpserver import RequestHandler
    from .httpserver import RequestMatcherKwargs
    from .httpserver import URIPattern


class BakedHTTPServer:
    """
    A proxy for :py:class:`HTTPServer` with pre-configured defaults for
    ``expect_request()`` and related methods.

    Created via :py:meth:`HTTPServer.bake`. Keyword arguments stored at bake
    time are merged with arguments provided at call time using last-wins
    semantics: if the same keyword appears in both, the call-time value is
    used.

    Any attribute not explicitly defined here is delegated to the wrapped
    :py:class:`HTTPServer`, so ``url_for()``, ``check_assertions()``, etc.
    work transparently.
    """

    def __init__(self, server: HTTPServer, **kwargs: Unpack[RequestMatcherKwargs]) -> None:
        self._server = server
        self._defaults = kwargs
        self._context_depth: int = 0
        self._started_server: bool = False

    def __enter__(self) -> Self:
        if self._context_depth == 0:
            self._started_server = not self._server.is_running()
            self._server.__enter__()
        self._context_depth += 1
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._context_depth -= 1
        if self._started_server and self._context_depth == 0:
            self._server.__exit__(exc_type, exc_value, traceback)
            self._started_server = False

    def __getattr__(self, name: str) -> Any:
        return getattr(self._server, name)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} defaults={self._defaults!r} server={self._server!r}>"

    def _merge_kwargs(self, kwargs: RequestMatcherKwargs) -> RequestMatcherKwargs:
        return self._defaults | kwargs

    def bake(self, **kwargs: Unpack[RequestMatcherKwargs]) -> Self:
        """
        Create a new :py:class:`BakedHTTPServer` by further layering defaults.

        The new proxy merges the current defaults with the new ``kwargs``.
        """
        return self.__class__(self._server, **self._merge_kwargs(kwargs))

    def expect_request(
        self,
        uri: str | URIPattern | Pattern[str],
        **kwargs: Unpack[RequestMatcherKwargs],
    ) -> RequestHandler:
        """Create and register a request handler, using baked defaults."""
        return self._server.expect_request(uri, **self._merge_kwargs(kwargs))

    def expect_oneshot_request(
        self,
        uri: str | URIPattern | Pattern[str],
        **kwargs: Unpack[RequestMatcherKwargs],
    ) -> RequestHandler:
        """Create and register a oneshot request handler, using baked defaults."""
        return self._server.expect_oneshot_request(uri, **self._merge_kwargs(kwargs))

    def expect_ordered_request(
        self,
        uri: str | URIPattern | Pattern[str],
        **kwargs: Unpack[RequestMatcherKwargs],
    ) -> RequestHandler:
        """Create and register an ordered request handler, using baked defaults."""
        return self._server.expect_ordered_request(uri, **self._merge_kwargs(kwargs))



"""
Hooks for pytest-httpserver
"""

import os
import time
from collections.abc import Callable

from werkzeug import Request
from werkzeug import Response


class Chain:
    """
    Combine multiple hooks into one callable object

    Hooks specified will be called one by one.

    Each hook will receive the response object made by the previous hook,
    similar to reduce.
    """

    def __init__(self, *args: Callable[[Request, Response], Response]) -> None:
        """
        :param *args: callable objects specified in the same order they should
            be called.
        """
        self._hooks = args

    def __call__(self, request: Request, response: Response) -> Response:
        """
        Calls the callable object one by one. The second and further callable
        objects receive the response returned by the previous one, while the
        first one receives the original response object.
        """
        for hook in self._hooks:
            response = hook(request, response)
        return response


class Delay:
    """
    Delays returning the response
    """

    def __init__(self, seconds: float) -> None:
        """
        :param seconds: seconds to sleep before returning the response
        """
        self._seconds = seconds

    def _sleep(self) -> None:
        """
        Sleeps for the seconds specified in the constructor
        """
        time.sleep(self._seconds)

    def __call__(self, _request: Request, response: Response) -> Response:
        """
        Delays returning the response object for the time specified in the
        constructor. Returns the original response unmodified.
        """
        self._sleep()
        return response


class Garbage:
    def __init__(self, prefix_size: int = 0, suffix_size: int = 0) -> None:
        """
        Adds random bytes to the beginning or to the end of the response data.

        :param prefix_size: amount of random bytes to be added to the beginning
            of the response data

        :param suffix_size: amount of random bytes to be added to the end
            of the response data

        """
        assert prefix_size >= 0, "prefix_size should be positive integer"
        assert suffix_size >= 0, "suffix_size should be positive integer"
        self._prefix_size = prefix_size
        self._suffix_size = suffix_size

    def _get_garbage_bytes(self, size: int) -> bytes:
        """
        Returns the specified amount of random bytes.

        :param size: amount of bytes to return
        """
        return os.urandom(size)

    def __call__(self, _request: Request, response: Response) -> Response:
        """
        Adds random bytes to the beginning or to the end of the response data.

        New random bytes will be generated for every call.

        Returns the modified response object.
        """
        prefix = self._get_garbage_bytes(self._prefix_size)
        suffix = self._get_garbage_bytes(self._suffix_size)
        response.set_data(prefix + response.get_data() + suffix)
        return response



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import pytest

from .httpserver import HTTPServer

if TYPE_CHECKING:
    from collections.abc import Generator
    from ssl import SSLContext


class Plugin:
    SERVER: PluginHTTPServer | None = None


class PluginHTTPServer(HTTPServer):
    def start(self) -> None:
        super().start()
        Plugin.SERVER = self

    def stop(self) -> None:
        super().stop()
        Plugin.SERVER = None


def get_httpserver_listen_address() -> tuple[str | None, int | None]:
    listen_host = os.environ.get("PYTEST_HTTPSERVER_HOST")
    listen_port_str = os.environ.get("PYTEST_HTTPSERVER_PORT")
    listen_port: int | None = int(listen_port_str) if listen_port_str else None

    return listen_host, listen_port


@pytest.fixture(scope="session")
def httpserver_listen_address() -> tuple[str | None, int | None]:
    return get_httpserver_listen_address()


@pytest.fixture(scope="session")
def httpserver_ssl_context() -> None:
    return None


@pytest.fixture(scope="session")
def make_httpserver(
    httpserver_listen_address: tuple[str | None, int | None],
    httpserver_ssl_context: SSLContext | None,
) -> Generator[HTTPServer, None, None]:
    host, port = httpserver_listen_address
    if not host:
        host = HTTPServer.DEFAULT_LISTEN_HOST
    if not port:
        port = HTTPServer.DEFAULT_LISTEN_PORT

    server = HTTPServer(host=host, port=port, ssl_context=httpserver_ssl_context)
    server.start()
    yield server
    server.clear()
    if server.is_running():
        server.stop()


def pytest_sessionfinish(session: pytest.Session, exitstatus: int) -> None:  # noqa: ARG001
    if Plugin.SERVER is not None:
        Plugin.SERVER.clear()
        if Plugin.SERVER.is_running():
            Plugin.SERVER.stop()


@pytest.fixture
def httpserver(make_httpserver: HTTPServer) -> HTTPServer:
    server = make_httpserver
    server.clear()
    return server


@pytest.fixture(scope="session")
def make_httpserver_ipv4(
    httpserver_ssl_context: SSLContext | None,
) -> Generator[HTTPServer, None, None]:
    server = HTTPServer(host="127.0.0.1", port=0, ssl_context=httpserver_ssl_context)
    server.start()
    yield server
    server.clear()
    if server.is_running():
        server.stop()


@pytest.fixture
def httpserver_ipv4(make_httpserver_ipv4: HTTPServer) -> HTTPServer:
    server = make_httpserver_ipv4
    server.clear()
    return server


@pytest.fixture(scope="session")
def make_httpserver_ipv6(
    httpserver_ssl_context: SSLContext | None,
) -> Generator[HTTPServer, None, None]:
    server = HTTPServer(host="::1", port=0, ssl_context=httpserver_ssl_context)
    server.start()
    yield server
    server.clear()
    if server.is_running():
        server.stop()


@pytest.fixture
def httpserver_ipv6(make_httpserver_ipv6: HTTPServer) -> HTTPServer:
    server = make_httpserver_ipv6
    server.clear()
    return server
